@@ -25,17 +25,15 @@ import unicodedata
 
 
 class Parser:
-    def __init__(self, parms):
+    def __init__(self, parms, read_macros=None):
         self.parms = parms
+        self.read_macros = read_macros
         self.the_macros = dict((m.name, m) for m in parms.macro_defs_python)
         self.the_environments = dict((e.name, e)
                                             for e in parms.environment_defs)
         self.mathparser = mathparser.MathParser(self)
         self.unknowns = []
-
-        # read macro definitions from LaTeX text
-        macs = parms.scanner.scan(parms.macro_defs_latex)
-        self.expand_sequence(scanner.Buffer(macs))
+        self.latex = ''
 
         # used by expand_item():
         self.item_macro = defs.Macro(parms, '\\item', args='O', repl='#1')
@@ -43,8 +41,8 @@ class Parser:
         # for savety: \item labels outside of any environment
         def labs_default(level):
             while True:
-                yield parms.item_default_label
-        self.item_lab_stack = [(labs_default(1), '')]
+                yield parms.item_default_label[0]
+        self.item_lab_stack = [(labs_default(0), '')]
 
         # for expansion of verbatim environment
         self.verbatim_begin = [
@@ -55,15 +53,30 @@ class Parser:
         ]
         self.verbatim_end = parms.scanner.scan('\\end{verbatim}')
 
+        # read macro definitions from LaTeX text
+        self.parser_work(parms.macro_defs_latex)
+
     #   scan and parse (expand) LaTeX string to tokens
     #
-    def parse(self, latex, extract=None):
+    def parser_work(self, latex):
+        # save self.latex for nested calls, e.g. \LTmacros{}
+        latex_sav = self.latex
+        self.latex = latex
         toks = self.parms.scanner.scan(latex)
+        toks = self.expand_sequence(scanner.Buffer(toks))
+        self.latex = latex_sav
+        return toks
+
+    #   main entry point
+    #
+    def parse(self, latex, extract=None):
         if extract:
             self.init_extractions(extract)
         self.extracted = []
         self.unknowns = []
-        main = self.expand_sequence(scanner.Buffer(toks))
+
+        main = self.parser_work(latex)
+
         if extract:
             main = []
         for extr in self.extracted:
@@ -132,9 +145,11 @@ class Parser:
                                                             tok.environ)
                 continue
             elif tok.txt == '$$' or tok.txt == '\\[':
-                assert self.parms.math_default_env in self.the_environments
+                if self.parms.math_default_env not in self.the_environments:
+                    utils.fatal('no environment for \'$$\' or \'\\[\'')
                 env = self.the_environments[self.parms.math_default_env]
-                assert type(env) is defs.EquEnv
+                if type(env) is not defs.EquEnv:
+                    utils.fatal(repr(env.name) + ' is not an EquEnv')
                 out += self.mathparser.expand_display_math(buf, tok, env)
                 continue
             elif type(tok) is defs.AccentToken:
@@ -169,11 +184,10 @@ class Parser:
     #   - buffer will contain at least one token for position tracking
     #   - this also ensures that an empty option [] will be "something"
     #
-    def arg_buffer(self, buf, end='}'):
+    def arg_buffer(self, buf, start, end='}'):
         tok = buf.skip_space()
         if not tok:
-            # XXX: improve position estimation?
-            return scanner.Buffer([defs.VoidToken(0)])
+            return scanner.Buffer([defs.VoidToken(start)])
         if type(tok) is defs.ParagraphToken:
             return scanner.Buffer([defs.VoidToken(tok.pos)])
         if end == '}' and tok.txt != '{':
@@ -196,7 +210,8 @@ class Parser:
                 return scanner.Buffer(out)
             out.append(tok)
             tok = buf.next()
-        utils.latex_error('cannot find closing ' + end, pos)
+        return scanner.Buffer(utils.latex_error('cannot find closing ' + end,
+                                            pos, self.latex, self.parms) + out)
 
     #   expand a "normal" macro
     #   Return: tokens to be inserted
@@ -216,23 +231,31 @@ class Parser:
     def expand_arguments(self, buf, mac, start):
         arguments = []
         arguments_extr = []
+        pos = start
         for n, code in enumerate(mac.args):
             arg_extr = arg = []
             tok = buf.skip_space()
+            if tok:
+                pos = tok.pos
             if code == '*':
                 if tok and tok.txt == '*':
                     arg_extr = arg = [tok]
                     buf.next()
             elif code == 'O':
                 if tok and tok.txt == '[':
-                    arg_extr = arg = self.arg_buffer(buf, end=']').all()
+                    arg_extr = arg = self.arg_buffer(buf, pos, end=']').all()
                 else:
-                    if n < len(mac.opts):
-                        arg = mac.opts[n]
+                    if n < len(mac.defaults):
+                        # NB: do not use positions from macro definition
+                        arg = [copy.copy(t) for t in mac.defaults[n]]
+                        for t in arg:
+                            t.pos = pos
+                            t.pos_fix = True
             elif code == 'A':
-                arg_extr = arg = self.arg_buffer(buf).all()
+                arg_extr = arg = self.arg_buffer(buf, pos).all()
             else:
-                assert code in '*AO'
+                utils.fatal('illegal arg code ' + repr(code)
+                                + ' of ' + repr(mac.name))
             arguments.append(arg)
             arguments_extr.append(arg_extr)
 
@@ -274,7 +297,7 @@ class Parser:
 
     def expand_accent(self, buf, tok):
         buf.next()
-        args = self.expand_sequence(self.arg_buffer(buf))
+        args = self.expand_sequence(self.arg_buffer(buf, tok.pos))
         if not args or not args[0].txt:
             c = ''
         elif len(args[0].txt) == 1:
@@ -289,29 +312,30 @@ class Parser:
             c = ' '.join(self.parms.accent_macros[tok.txt])
         else:
             if not ('a' <= c <= 'z' or 'A' <= c <= 'Z'):
-                utils.latex_error('text-mode accent for non-letter', tok.pos)
+                return utils.latex_error('text-mode accent for non-letter',
+                                            tok.pos, self.latex, self.parms)
             c = ('LATIN ' + ('SMALL' if c.islower() else 'CAPITAL')
                         + ' LETTER ' + c.upper() + ' WITH ' 
                         + self.parms.accent_macros[tok.txt][0])
         try:
             u = unicodedata.lookup(c)
         except:
-            utils.latex_error('could not find UTF-8 character "'
-                                            + c + '"', tok.pos)
+            return utils.latex_error('could not find UTF-8 character "' + c
+                                    + '"', tok.pos, self.latex, self.parms)
         return [defs.TextToken(tok.pos, u)] + args
 
     #   open an environment
     #
     def begin_environment(self, buf, tok, math):
         out = [defs.ActionToken(tok.pos)]
-        name = self.get_environment_name(buf)
+        name = self.get_environment_name(buf, tok)
         if name not in self.the_environments:
             if not (math or name in self.unknowns):
                 self.unknowns.append(name)
             return out
         env = self.the_environments[name]
         if env.items:
-            level = len([v for v in self.item_lab_stack if v[1] == name]) + 1
+            level = len([v for v in self.item_lab_stack if v[1] == name])
             self.item_lab_stack.append((env.items(level), name))
         if env.add_pars:
             out = [defs.ParagraphToken(tok.pos, '\n\n', pos_fix=True)]
@@ -327,7 +351,7 @@ class Parser:
     #   - second element of returned 2-tuple: reached env_stop
     #
     def end_environment(self, buf, tok, env_stop):
-        name = self.get_environment_name(buf)
+        name = self.get_environment_name(buf, tok)
         out = [defs.ActionToken(tok.pos)]
         if name in self.the_environments:
             env = self.the_environments[name]
@@ -337,9 +361,9 @@ class Parser:
                 out = [defs.ParagraphToken(tok.pos, '\n\n', pos_fix=True)]
         return out, name == env_stop
 
-    def get_environment_name(self, buf):
+    def get_environment_name(self, buf, tok):
         buf.next()
-        return self.get_text_expanded(self.arg_buffer(buf).all())
+        return self.get_text_expanded(self.arg_buffer(buf, tok.pos).all())
 
     #   generate string from token sequence, without macro expansion
     #
@@ -377,16 +401,18 @@ class Parser:
         tok.can_end = True
         tokens.append(tok)
 
+        # avoid modifications at list begin (expensive for long lists)
+        tokens = list(reversed(tokens))
         out = []
         while tokens:
-            tok = tokens.pop(0)
+            tok = tokens.pop()
             if not tok.can_start:
                 out.append(tok)
                 continue
             buf = [tok]
             can_remove = True
             while tokens:
-                tok = tokens.pop(0)
+                tok = tokens.pop()
                 buf.append(tok)
                 if tok.can_end:
                     break
@@ -412,9 +438,14 @@ class Parser:
                 else:
                     t2.txt = ''
                     t2.pos += len(txt)
-                buf = [t1, t2]
-            if len(buf) > 1:
-                tokens.insert(0, eval(buf.pop()))
+                buf = [t1]
+                tokens.append(eval(t2))
+                # NB: we deleted a line break
+                tok = eval(defs.TextToken(t2.pos, ''))
+                tok.can_start = True
+                tokens.append(tok)
+            elif len(buf) > 1:
+                tokens.append(eval(buf.pop()))
             out += buf
 
         return [t for t in out if t.txt]
